import {desiredHeight, desiredWidth, getScreen} from "./screen.ts";
import {BG_OUTER} from "./colors.ts";
import {getInput} from "./input.ts";
import {checkGrid, ConceptualCell, maps, mapSzX, mapSzY} from "./maps.ts";

type Point = {x: number, y: number}

class MenuCamera {
  // measured in whole screens
  position: Point;
  target: Point

  constructor({position, target}: {position: Point, target: Point}) {
    this.position = position;
    this.target = target;
  }

  update() {
    let adjust = (x0: number, x1: number) => {
      if (Math.abs(x1 - x0) < 0.01) { return x1; }
      return (x0 * 8 + x1 * 2) / 10;
    }
    this.position.x = adjust(this.position.x, this.target.x);
    this.position.y = adjust(this.position.y, this.target.y);
  }
}

type GameState = "Gameplay" | "Thralls";

function getScreenLocation(state: GameState): {x: number, y: number} {
  if (state === "Gameplay") {
    return {x: 0.0, y: 0.0}
  }
  if (state === "Thralls") {
    return {x: 0.0, y: 1.0}
  }

  throw `invalid state: ${state}`
}

export class Game {
  camera: MenuCamera;
  state: GameState;
  huntMode: HuntMode;

  constructor() {
    this.camera = new MenuCamera({
      position: {x: 0.0, y: 0.0},
      target: {x: 0.0, y: 0.0}
    });
    this.state = "Gameplay";

    this.huntMode = HuntMode.generate({depth: 1});
  }

  update() {
    if (getInput().isPressed("w")) {
      this.state = "Gameplay"
    }
    if (getInput().isPressed("s")) {
      this.state = "Thralls"
    }

    this.camera.target = getScreenLocation(this.state);
    this.camera.update();

    // state-specific updates
    this.updateGameplay();
  }

  draw() {
    let screen = getScreen();
    let ctx = screen.makeContext();

    // draw screen background
    ctx.fillStyle = BG_OUTER;
    ctx.fillRect(0, 0, screen.w, screen.h);

    this.drawGameplay();

    // we draw all states at once and pan between them
    // mainFont.drawText({ctx: ctx, text: "You have been given a gift.", x: 0, y: 0})
    let xy = this.getCameraMouseXy();
    if (xy != null) {
      ctx = this.makeCameraContext();
      ctx.globalCompositeOperation = "difference";
      if (getInput().isDown("leftMouse")) {
        ctx.fillStyle = "#ff0";
      } else {
        ctx.fillStyle = "#fff";
      }
      ctx.fillRect(xy.x - 1, xy.y - 1, 3, 3);
      ctx.fillRect(xy.x, xy.y - 1, 1, 3);
      ctx.fillRect(xy.x - 1, xy.y, 3, 1);
    }
  }

  getCameraOffset(): Point {
    let screen = getScreen();
    let {w, h} = screen;
    return {
      x: Math.floor(w * this.camera.position.x),
      y: Math.floor(h * this.camera.position.y)
    };
  }

  getCameraMouseXy(): Point | null {
    let xy = getInput().mouseXy();
    if (xy == null) {
      return null;
    }
    let {x: dx, y: dy} = this.getCameraOffset();
    return {
      x: xy.x + dx,
      y: xy.y + dy,
    };
  }

  makeCameraContext() {
    let screen = getScreen();
    let ctx = screen.makeContext();
    let {x, y} = this.getCameraOffset();

    ctx.translate(-x, -y);
    return ctx;
  }

  getPaneRegionForGameState(gameState: GameState) {
    let screen = getScreen();
    let {w, h} = screen;
    let overallScreenLocation = getScreenLocation(gameState);

    let bigPaneX = overallScreenLocation.x * w;
    let bigPaneY = overallScreenLocation.y * h;
    let bigPaneW = w;
    let bigPaneH = h;

    let smallPaneW = desiredWidth;
    let smallPaneH = desiredHeight;
    let smallPaneX = Math.floor(bigPaneX + (bigPaneW - smallPaneW) / 2)
    let smallPaneY = Math.floor(bigPaneY + (bigPaneH - smallPaneH) / 2)

    return {
      big: {
        position: {x: bigPaneX, y: bigPaneY},
        size: {x: bigPaneW, y: bigPaneH}
      },
      small: {
        position: {x: smallPaneX, y: smallPaneY},
        size: {x: smallPaneW, y: smallPaneH}
      }
    }

  }

  updateGameplay() {

  }

  drawGameplay() {
    let region = this.getPaneRegionForGameState("Gameplay")
    let ctx = this.makeCameraContext()
    ctx.translate(region.small.position.x, region.small.position.y)


  }
}

type Stat = "AGI" | "INT" | "CHA" | "PSI";
const ALL_STATS: Array<Stat> = ["AGI", "INT", "CHA", "PSI"];
type MapCellContent =
  {type: "statPickup", stat: Stat} |
  {type: "stairs"} |
  {type: "block"}

type MapCell = {
  content: MapCellContent,
  revealed: boolean
}

class HuntMode {
  depth: number
  cells: Array<Array<MapCell>>
  player: Point | null

  constructor({depth, cells, player}: {depth: number, cells: Array<Array<MapCell>>, player: Point | null}) {
    this.depth = depth;
    this.cells = cells;
    this.player = player;

    checkGrid(this.cells);
  }

  static generate({depth}: {depth: number}) {
    let mapNames: Array<string> = Object.keys(maps);
    let mapName = mapNames[Math.floor(Math.random() * mapNames.length)];
    let map = maps[mapName];

    let rows = [];
    for (let y = 0; y < mapSzY; y++) {
      let row = [];
      for (let x = 0; x < mapSzX; x++) {
        let src = map[y][x];
        row.push(HuntMode.#generateCell(src))
      }
      rows.push(row);
    }

    if (Math.random() < 0.75) {
      while (true) {
        let x = Math.floor(Math.random() * mapSzX);
        let y = Math.floor(Math.random() * mapSzY);

        let item = rows[y][x];
        if (item.content.type != "block") {
          item.content = {type: "stairs"}
          break;
        }
      }
    }

    return new HuntMode({depth, cells: rows, player: null})
  }

  static #generateCell(conceptual: ConceptualCell): MapCell {
    switch (conceptual) {
      case "X":
        return { content: {type: "block"}, revealed: true};
      case " ":
        return { content: HuntMode.#generateContent(true), revealed: true };
      case ".":
        return { content: HuntMode.#generateContent(false), revealed: true };
    }
  }

  static #generateContent(_revealed: boolean): MapCellContent {
    // stat pickup
    let gsp = (): MapCellContent => {
      return {type: "statPickup", stat: choose(ALL_STATS)}
    };
    // TODO: Other objects?
    return choose([
      gsp, gsp, gsp, gsp
    ])();
  }

  get(at: Point) {
  }
}

function choose<T>(array: Array<T>): T  {
  if (array.length == 0) {
    throw `array cannot have length 0 for choose`
  }
  return array[Math.floor(Math.random() * array.length)]
}

export let game = new Game();